def test_us_prototype():
    from policyengine_us import Microsimulation, Simulation
    import numpy as np
    import torch
    from torch.utils.tensorboard import SummaryWriter

    # Initialize a TensorBoard writer
    writer = SummaryWriter()

    # Create a Microsimulation instance
    sim = Microsimulation()

    # Compute income and payroll taxes. These are MicroSeries objects from the microdf library
    income_tax_microseries = sim.calculate(
        "income_tax", 2023, map_to="household"
    )
    payroll_tax_microseries = sim.calculate(
        "employee_payroll_tax", 2023, map_to="household"
    )

    # Convert them into usable NumPy arrays
    sim_income_tax = np.array(income_tax_microseries)
    sim_payroll_tax = np.array(payroll_tax_microseries)
    sim_weights = np.array(income_tax_microseries.weights)

    log_weights = np.log(sim_weights)

    # Initialize usable ground truth income and payroll tax values
    targets = np.array([2_176_000_000_000, 1_614_454_000_000])
    target_names = np.array(["income tax revenue", "payroll tax revenue"])

    sim_matrix = np.array([sim_income_tax, sim_payroll_tax])

    # sim_matrix (cross) exp(log_weights) = targets
    log_weights = torch.tensor(
        log_weights, dtype=torch.float32, requires_grad=True
    )
    sim_matrix = torch.tensor(sim_matrix, dtype=torch.float32)
    targets = torch.tensor(targets, dtype=torch.float32)

    optimizer = torch.optim.Adam([log_weights])

    # Training loop
    num_epochs = 20000
    for epoch in range(num_epochs):

        # Estimate the targets
        targets_estimate = sim_matrix @ torch.exp(log_weights)
        # Calculate the loss
        loss = torch.mean((targets_estimate - targets) ** 2)

        writer.add_scalar("Loss/train", loss, epoch)

        optimizer.zero_grad()

        # Perform backpropagation
        loss.backward()

        # Update weights
        optimizer.step()

        # Print loss for every 1000 epochs
        if epoch % 1000 == 0:
            print(f"Epoch {epoch}, Loss: {loss.item()}")

    writer.flush()

    final_weights = np.exp(log_weights.detach().numpy())
    final_estimates = (
        np.array([sim_income_tax, sim_payroll_tax]) @ final_weights
    )
    true_values = targets
    print("Final weights:", final_weights)
    print("Final estimates:", final_estimates)
    print("True values:", true_values)


def test_us_microsimulation():
    from policyengine_us import Microsimulation

    # Create a Microsimulation instance
    sim = Microsimulation()


def test_us_reweight():
    from policyengine_us import Microsimulation
    from reweight import reweight
    import torch

    sim = Microsimulation()

    from policyengine_us.data.datasets.cps.enhanced_cps.loss import (
        generate_model_variables,
    )

    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_model_variables("cps_2021", 2025)

    sim_matrix = torch.tensor(values_df.to_numpy(), dtype=torch.float32)
    weights_tensor = torch.tensor(household_weights, dtype=torch.float32)
    targets_tensor = torch.tensor(targets_array, dtype=torch.float32)
    reweight(weights_tensor, sim_matrix, targets, targets_tensor)



def test_uk_microsimulation():
    from policyengine_uk import Microsimulation

    # Create a Microsimulation instance
    sim = Microsimulation()


def test_uk_reweight():
    from policyengine_uk import Microsimulation
    from reweight import reweight
    import torch

    sim = Microsimulation()

    from policyengine_uk.data import RawFRS_2021_22

    RawFRS_2021_22().download()

    from policyengine_uk.data.datasets.frs.calibration.calibrate import (
        generate_model_variables,
    )

    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_model_variables("frs_2021", 2025)

    sim_matrix = torch.tensor(values_df.to_numpy(), dtype=torch.float32)
    reweight(household_weights, sim_matrix, targets, targets_array)



def test_imports():
    try:
        import torch
        import numpy as np
        import policyengine_core
        from torch.utils.tensorboard import SummaryWriter
    except:
        raise AssertionError("Failed to import necessary libraries")


def test_install():
    try:
        import reweight
    except:
        raise AssertionError("Failed to build reweight")


def test_secret_usage():
    import os

    token = os.environ["POLICYENGINE_GITHUB_MICRODATA_AUTH_TOKEN"]
    token_not_none = token != None
    assert token_not_none, "Authentication token is None"



"""This file contains metadata and dependencies."""

from setuptools import find_packages, setup

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

setup(
    name="reweight",
    version="0.3.0",
    author="PolicyEngine",
    author_email="hello@policyengine.org",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 1 - Planning",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    description="Reweighting library for survey data for use with PolicyEngine",
    keywords="benefit microsimulation social tax",
    license="http://www.fsf.org/licensing/licenses/agpl-3.0.html",
    url="https://github.com/PolicyEngine/reweight",
    include_package_data=True,  # Will read MANIFEST.in
    install_requires=[
        "numpy<2.0",
        "pandas",
        "torch",
        "tensorboard",
        "jupyter-book",
        "pytest",
        "policyengine-core~=2.21.8",
        "policyengine-us~=0.794.1",
        "policyengine-uk",
        "survey_enhance",
    ],
    extras_require={
        "dev": [
            "black",
            "yaml-changelog",
            "setuptools",
            "setuptools_scm",
        ],
    },
    # Windows CI requires Python 3.9.
    python_requires=">=3.7",
    packages=find_packages(),
)



def test_integrated_prototype():
    import numpy as np
    import torch
    from torch.utils.tensorboard import SummaryWriter

    # Initialize a TensorBoard SummaryWriter instance
    writer = SummaryWriter()

    # Survey of three records

    # Three households surveyed, survey gave them weights of 100, 500, 300
    initial_weights = np.array([100, 500, 300])
    log_weights = np.log(initial_weights)

    # total population, income tax revenue, VAT revenue
    targets = np.array([1e6, 5e9, 6e9])
    target_names = np.array(
        ["total population", "income tax revenue", "VAT revenue"]
    )

    # initial weights (cross) matrix = targets
    matrix = np.array([[1, 1, 1], [300, 500, 250], [100, 200, 300]])

    log_weights = torch.tensor(
        log_weights, dtype=torch.float32, requires_grad=True
    )
    matrix = torch.tensor(matrix, dtype=torch.float32)
    targets = torch.tensor(targets, dtype=torch.float32)

    optimizer = torch.optim.Adam([log_weights])

    # Training loop
    num_epochs = 1_000
    for epoch in range(num_epochs):

        # Estimate the targets
        targets_estimate = matrix @ torch.exp(log_weights)
        # Calculate the loss
        loss = torch.mean((targets_estimate - targets) ** 2)

        writer.add_scalar("Loss/train", loss, epoch)

        optimizer.zero_grad()

        # Perform backpropagation
        loss.backward()

        # Update weights
        optimizer.step()

        # Print loss for every 1000 epochs
        if epoch % 100 == 0:
            print(f"Epoch {epoch}, Loss: {loss.item()}")

    writer.flush()

    # Print final weights
    print("Final weights:", np.exp(log_weights.detach().numpy()))



import pandas as pd
import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter


def reweight(
    initial_weights,
    estimate_matrix,
    target_names,
    target_values,
    epochs=1000,
    epoch_step=100,
):
    """
    Main reweighting function, suitable for PolicyEngine UK use (PolicyEngine US use and testing TK)

    To avoid the need for equivalisation factors, use relative error:
    |predicted - actual|/actual

    Parameters:
    household_weights (torch.Tensor): The initial weights given to survey data, which are to be
    adjusted by this function.
    estimate_matrix (torch.Tensor): A large matrix of estimates, obtained from e.g. a PolicyEngine
    Microsimulation instance.
    target_names (iterable): The names of a set of target statistics treated as ground truth.
    target_values (torch.Tensor): The values of these target statistics.
    epochs: The number of iterations that the optimization loop should run for.
    epoch_step: The interval at which to print the loss during the optimization loop.

    Returns:
    final_weights: a reweighted set of household weights, obtained through an optimization process
    over mean squared errors with respect to the target values.
    """
    # Initialize a TensorBoard writer
    writer = SummaryWriter()

    # Create a Torch tensor of log weights
    log_weights = torch.log(initial_weights)
    log_weights.requires_grad_()

    # estimate_matrix (cross) exp(log_weights) = target_values

    optimizer = torch.optim.Adam([log_weights])

    # Report the initial loss:
    targets_estimate = torch.exp(log_weights) @ estimate_matrix
    # Calculate the loss
    loss = torch.mean(
        ((targets_estimate - target_values) / target_values) ** 2
    )
    print(f"Initial loss: {loss.item()}")

    # Training loop
    for epoch in range(epochs):

        # Estimate the targets
        targets_estimate = torch.exp(log_weights) @ estimate_matrix
        # Calculate the loss
        loss = torch.mean(
            ((targets_estimate - target_values) / target_values) ** 2
        )

        writer.add_scalar("Loss/train", loss, epoch)

        optimizer.zero_grad()

        # Perform backpropagation
        loss.backward()

        # Update weights
        optimizer.step()

        # Print loss whenever the epoch number, when one-indexed, is divisible by epoch_step
        if (epoch + 1) % epoch_step == 0:
            print(f"Epoch {epoch+1}, Loss: {loss.item()}")

    writer.flush()

    return torch.exp(log_weights.detach())



import pandas as pd
import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter
import os
import requests
import base64

import policyengine_uk
from policyengine_uk import Microsimulation
from policyengine_uk.data import RawFRS_2021_22
from policyengine_uk.data.datasets.frs.calibration.calibrate import (
    generate_model_variables as uk_generate,
)

import policyengine_us
from policyengine_us.data.datasets.cps.enhanced_cps.loss import (
    generate_model_variables as us_generate,
)

from reweight import reweight


def generate_country_weights(year, data_source, generate_func):
    """
    Parameters:
    year (int): The year for which these country values are generated.
    data_source (str): The name of the data source for that country.
    generate_func (function): The function used to generate the initial values.

    Returns:
    final_weights (torch.Tensor): a PyTorch tensor of final reweighted weights.
    """
    (
        household_weights,
        weight_adjustment,
        values_df,
        targets,
        targets_array,
        equivalisation_factors_array,
    ) = generate_func(data_source, year)
    sim_matrix = torch.tensor(values_df.to_numpy(), dtype=torch.float32)
    initial_weights = torch.tensor(household_weights, dtype=torch.float32)
    targets_tensor = torch.tensor(targets_array, dtype=torch.float32)
    final_weights = reweight(
        initial_weights, sim_matrix, targets, targets_tensor, epochs=1_000
    )
    return final_weights


def generate_country_csv(
    start_year, end_year, data_source, generate_func, csv_filename
):
    """
    Parameters:
    start_year (int): The year for which these country values start generating (inclusive).
    end_year (int): The year for which these country values stop generating (non-inclusive).
    data_source (str): The name of the data source for that country.
    generate_func (function): The function used to generate the initial values.
    csv_filename (str): The name of the file which the generated data are saved under.

    Returns:
    None. Generates and saves a CSV file of reweighted weights.
    """
    weights_df = pd.DataFrame()
    for year in range(start_year, end_year):
        final_weights = generate_country_weights(
            year, data_source, generate_func
        )
        weight_series = pd.Series(final_weights.numpy())
        weights_df[str(year)] = weight_series
    weights_df.to_csv(csv_filename)


RawFRS_2021_22().download()
generate_country_csv(
    2024, 2029, "frs_2021", uk_generate, "updated_uk_weights.csv"
)
generate_country_csv(
    2024, 2029, "cps_2021", us_generate, "updated_us_weights.csv"
)

# Now, create a GitHub release

api_url = "https://api.github.com/repos/PolicyEngine/reweight/releases"

owner = "pmberg"
repo = "reweight"
token = os.environ.get("API_GITHUB_TOKEN")

# Create release
headers = {
    "Authorization": f"token {token}",
    "Accept": "application/vnd.github.v3+json",
}
release_data = {
    "tag_name": f'v{pd.Timestamp.now().strftime("%Y.%m.%d.%H.%M.%S")}',
    "name": f'Data Release {pd.Timestamp.now().strftime("%Y.%m.%d.%H.%M.%S")}',
    "body": "Automated data release with updated weights",
}
response = requests.post(
    api_url.format(owner=owner, repo=repo), headers=headers, json=release_data
)
release = response.json()
print(release)
# Upload assets
upload_url = release["upload_url"].split("{")[0]


def upload_file(file_name):
    with open(file_name, "rb") as file:
        content = file.read()
    headers["Content-Type"] = (
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    )
    params = {"name": os.path.basename(file_name)}
    response = requests.post(
        upload_url, headers=headers, params=params, data=content
    )
    if response.status_code == 201:
        print(f"File added successfully: {release['html_url']}")
    else:
        print(f"Failed to add file: {response.content}")


for file_name in ["updated_uk_weights.csv", "updated_us_weights.csv"]:
    upload_file(file_name)
